"""
A set of helpers to simplify the creation of MongoDB queries.
"""

from typing import Any

__all__ = [
    # Queries
    "Q",
]


# Queries
class Condition:
    """
    A query condition of the form `{path: {operator: value}}`.
    """

    def __init__(self, q, value, operator):
        self.q = q
        self.value = value
        self.operator = operator

    def to_dict(self) -> dict[str, dict[str, Any]]:
        """Return a dictionary suitable for use with pymongo as a filter"""
        if self.operator == "$eq":
            return {self.q: self.value}
        if self.q is None:
            return {self.operator: self.value}
        return {self.q: {self.operator: self.value}}


class QMeta(type):
    """
    Meta-class for query builder.
    """

    def __getattr__(self, name):
        return Q(name)

    def __getitem__(self, name):
        return Q(name)

    def __eq__(self, other):
        return Condition(None, other, "$eq")

    def __ge__(self, other):
        return Condition(None, other, "$gte")

    def __gt__(self, other):
        return Condition(None, other, "$gt")

    def __le__(self, other):
        return Condition(None, other, "$lte")

    def __lt__(self, other):
        return Condition(None, other, "$lt")

    def __ne__(self, other):
        return Condition(None, other, "$ne")


class Q(metaclass=QMeta):
    """
    Start point for the query creation, the Q class is a special type of class
    that's typically initialized by appending an attribute, for example:

        Q.hit_points > 100

    """

    def __init__(self, path):
        self._path = path

    def __eq__(self, other):
        return Condition(self._path, other, "$eq")

    def __ge__(self, other):
        return Condition(self._path, other, "$gte")

    def __gt__(self, other):
        return Condition(self._path, other, "$gt")

    def __le__(self, other):
        return Condition(self._path, other, "$lte")

    def __lt__(self, other):
        return Condition(self._path, other, "$lt")

    def __ne__(self, other):
        return Condition(self._path, other, "$ne")

    def __getattr__(self, name):
        self._path = "{0}.{1}".format(self._path, name)
        return self

    def __getitem__(self, name):
        self._path = "{0}.{1}".format(self._path, name)
        return self


class Group:
    """
    The Group class is used as a base class for operators that group together
    two or more conditions.
    """

    operator = ""

    def __init__(self, *conditions):
        self.conditions = conditions

    def to_dict(self):
        """Return a dictionary suitable for use with pymongo as a filter"""
        raw_conditions = []
        for condition in self.conditions:
            if isinstance(condition, (Condition, Group)):
                raw_conditions.append(condition.to_dict())
            else:
                raw_conditions.append(condition)
        return {self.operator: raw_conditions}



from typing import Any, ClassVar, Optional, Union

import msgspec
from bson import ObjectId
from pymongo import MongoClient

from .base import SpecBase, SpecProtocol, SubSpecBase
from .empty import Empty, EmptyObject
from .se import MongoEncoder, mongo_dec_hook, mongo_enc_hook

__all__ = ["Spec", "SubSpec"]


class Spec(msgspec.Struct, SpecBase, kw_only=True):
    _id: Union[ObjectId, msgspec.UnsetType] = msgspec.field(name="_id", default=msgspec.UNSET)  # type: ignore[assignment]
    _empty_type: ClassVar[Any] = msgspec.UNSET

    def encode(self, **encode_kwargs: Any) -> bytes:
        return msgspec.json.encode(self, **encode_kwargs) if encode_kwargs else MongoEncoder.encode(self)

    def decode(self, data: Any, **decode_kwargs: Any) -> Any:
        return msgspec.json.decode(data, type=self.__class__, dec_hook=mongo_dec_hook, **decode_kwargs)

    def to_json_type(self) -> Any:
        return msgspec.to_builtins(self, enc_hook=mongo_enc_hook)

    def to_dict(self) -> dict[str, Any]:
        return msgspec.structs.asdict(self)

    def to_tuple(self) -> tuple[Any, ...]:
        return msgspec.structs.astuple(self)

    @classmethod
    def get_fields(cls) -> set[str]:
        return set(cls.__struct_fields__)

    # msgspec Struct includes these by default- so we need to override them
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self._id == other._id

    def __lt__(self, other: Any) -> Any:
        return self._id < other._id


class SubSpec(msgspec.Struct, SubSpecBase, kw_only=True, dict=True):
    _parent: ClassVar[Any] = Spec

    def get(self, name, default=None):  # -> Any:
        return self.to_dict().get(name, default)

    def to_dict(self) -> dict[str, Any]:
        return msgspec.structs.asdict(self)


class MsgspecAdapter(SpecProtocol):
    _id: Union[EmptyObject, ObjectId] = msgspec.field(default=Empty)
    _empty_type: ClassVar[Any] = msgspec.UNSET

    def __init__(self, **data: Any) -> None: ...


class AdapterBuilder:
    def __call__(
        self, obj: type[msgspec.Struct], *, collection: str, client: Optional[MongoClient] = None, **kwds: Any
    ) -> Any:
        class BuiltSpecAdapter(SpecBase, obj):  # type: ignore
            _id: Union[EmptyObject, ObjectId] = msgspec.field(default=Empty)
            _empty_type: ClassVar[Any] = msgspec.UNSET

            def encode(self, **encode_kwargs: Any) -> bytes:
                return msgspec.json.encode(self, **encode_kwargs) if encode_kwargs else MongoEncoder.encode(self)

            def decode(self, data: Any, **decode_kwargs: Any) -> Any:
                return msgspec.json.decode(data, type=self.__class__, dec_hook=mongo_dec_hook, **decode_kwargs)

            def to_json_type(self) -> Any:
                return msgspec.to_builtins(self, enc_hook=mongo_enc_hook)

            def to_dict(self) -> dict[str, Any]:
                return msgspec.structs.asdict(self)

            def to_tuple(self) -> tuple[Any, ...]:
                return msgspec.structs.astuple(self)

            @classmethod
            def get_fields(cls) -> set[str]:
                return set(cls.__struct_fields__)

            # msgspec Struct includes these by default- so we need to override them
            def __eq__(self, other: Any) -> bool:
                if not isinstance(other, self.__class__):
                    return False

                return self._id == other._id

            def __lt__(self, other: Any) -> Any:
                return self._id < other._id

        BuiltSpecAdapter.__name__ = f"{obj.__name__}SpecAdapter"
        BuiltSpecAdapter._collection = collection
        BuiltSpecAdapter.__doc__ = obj.__doc__
        if client:
            BuiltSpecAdapter._client = client
        return BuiltSpecAdapter


SpecAdapter = AdapterBuilder()



from contextlib import contextmanager
from copy import deepcopy
from typing import Any, Callable, ClassVar, Generator, Mapping, Optional, Protocol, Sequence, TypeVar, Union

from blinker import signal
from bson import ObjectId
from pymongo import MongoClient, UpdateOne
from pymongo.collection import Collection
from pymongo.database import Database
from typing_extensions import Self

from mongospecs.empty import Empty, EmptyObject

from .query import Condition, Group

Specs = Sequence["SpecBase"]
RawDocuments = Sequence[dict[str, Any]]
SpecsOrRawDocuments = Union[Specs, RawDocuments]

T = TypeVar("T")


class SpecProtocol(Protocol):
    _client: ClassVar[Optional[MongoClient]] = None
    _db: ClassVar[Optional[Database]] = None
    _collection: ClassVar[Optional[str]] = None
    _collection_context: ClassVar[Optional[Collection]] = None
    _default_projection: ClassVar[dict[str, Any]] = {}
    _empty_type: ClassVar[Any] = Empty
    _id: Union[EmptyObject, ObjectId]

    @classmethod
    def get_fields(cls) -> set[str]: ...

    @classmethod
    def from_document(cls, document: dict[str, Any]) -> Self: ...

    def get(self, name, default=None) -> Any: ...

    def encode(self, **encode_kwargs: Any) -> bytes: ...

    def decode(self, data: Any, **decode_kwargs: Any) -> Any: ...

    def to_json_type(self) -> dict[str, Any]: ...

    def to_dict(self) -> dict[str, Any]: ...

    def to_tuple(self) -> tuple[Any, ...]: ...

    # Operations
    def insert(self) -> None:
        """Insert this document"""
        ...

    def unset(self, *fields: Any) -> None:
        """Unset the given list of fields for this document."""
        ...

    def update(self, *fields: Any) -> None:
        """
        Update this document. Optionally a specific list of fields to update can
        be specified.
        """
        ...

    def upsert(self, *fields: Any) -> None:
        """
        Update or Insert this document depending on whether it exists or not.
        The presense of an `_id` value in the document is used to determine if
        the document exists.

        NOTE: This method is not the same as specifying the `upsert` flag when
        calling MongoDB. When called for a document with an `_id` value, this
        method will call the database to see if a record with that Id exists,
        if not it will call `insert`, if so it will call `update`. This
        operation is therefore not atomic and much slower than the equivalent
        MongoDB operation (due to the extra call).
        """
        ...

    def delete(self) -> None:
        """Delete this document"""
        ...

    @classmethod
    def find(cls, filter=None, **kwargs) -> list[Mapping[str, Any]]:
        """Return a list of documents matching the filter"""
        ...

    @classmethod
    def find_one(cls, filter=None, **kwargs) -> Mapping[str, Any]:
        """Return the first document matching the filter"""
        ...

    def reload(self, **kwargs):
        """Reload the document"""
        ...

    @classmethod
    def insert_many(cls, documents: SpecsOrRawDocuments) -> Specs:
        """Insert a list of documents"""
        ...

    @classmethod
    def update_many(cls, documents: SpecsOrRawDocuments, *fields: Any) -> None:
        """
        Update multiple documents. Optionally a specific list of fields to
        update can be specified.
        """
        ...

    @classmethod
    def unset_many(cls, documents: SpecsOrRawDocuments, *fields: Any) -> None:
        """Unset the given list of fields for given documents."""
        ...

    @classmethod
    def delete_many(cls, documents: SpecsOrRawDocuments) -> None:
        """Delete multiple documents"""
        ...

    # Querying

    @classmethod
    def by_id(cls, id, **kwargs) -> Optional[Self]:
        """Get a document by ID"""
        ...

    @classmethod
    def count(cls, filter=None, **kwargs) -> int:
        """Return a count of documents matching the filter"""
        ...

    @classmethod
    def ids(cls, filter=None, **kwargs) -> list[ObjectId]:
        """Return a list of Ids for documents matching the filter"""
        ...

    @classmethod
    def one(cls, filter=None, **kwargs) -> Optional[Self]:
        """Return the first spec object matching the filter"""
        ...

    @classmethod
    def many(cls, filter=None, **kwargs) -> list[Self]:
        """Return a list of spec objects matching the filter"""
        ...

    @classmethod
    def get_collection(cls) -> Collection[Any]:
        """Return a reference to the database collection for the class"""
        ...

    @classmethod
    def get_db(cls) -> Database:
        """Return the database for the collection"""
        ...

    @classmethod
    @contextmanager
    def with_options(cls, **options: Any) -> Generator[Any, Any, None]: ...

    @classmethod
    def _path_to_value(cls, path: str, parent_dict: dict[str, Any]) -> Any:
        """Return a value from a dictionary at the given path"""
        ...

    @classmethod
    def _path_to_keys(cls, path: str) -> list[str]:
        """Return a list of keys for a given path"""
        ...

    @classmethod
    def _ensure_frames(cls, documents: SpecsOrRawDocuments) -> Specs:
        """
        Ensure all items in a list are frames by converting those that aren't.
        """
        ...

    @classmethod
    def _apply_sub_frames(cls, documents: RawDocuments, subs: dict[str, Any]) -> None:
        """Convert embedded documents to sub-frames for one or more documents"""
        ...

    @classmethod
    def _flatten_projection(cls, projection: dict[str, Any]) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
        """
        Flatten a structured projection (structure projections support for
        projections of (to be) dereferenced fields.
        """
        ...

    @classmethod
    def _dereference(cls, documents: RawDocuments, references: dict[str, Any]):
        """Dereference one or more documents"""
        ...

    # Signals
    @classmethod
    def listen(cls, event: str, func: Callable) -> None:
        """Add a callback for a signal against the class"""
        ...

    @classmethod
    def stop_listening(cls, event: str, func: Callable) -> None:
        """Remove a callback for a signal against the class"""
        ...

    # Integrity helpers

    @classmethod
    def cascade(cls, ref_cls, field, frames) -> None:
        """Apply a cascading delete (does not emit signals)"""
        ...

    @classmethod
    def nullify(cls, ref_cls, field, frames) -> None:
        """Nullify a reference field (does not emit signals)"""
        ...

    @classmethod
    def pull(cls, ref_cls, field, frames) -> None:
        """Pull references from a list field (does not emit signals)"""
        ...

    def __eq__(self, other: Any) -> bool: ...

    def __lt__(self, other: Any) -> Any: ...


class SpecBase:
    _client: ClassVar[Optional[MongoClient]] = None
    _db: ClassVar[Optional[Database]] = None
    _collection: ClassVar[Optional[str]] = None
    _collection_context: ClassVar[Optional[Collection]] = None
    _default_projection: ClassVar[dict[str, Any]] = {}
    _empty_type: ClassVar[Any] = Empty
    _id: Union[EmptyObject, ObjectId]

    @classmethod
    def get_fields(cls) -> set[str]:
        raise NotImplementedError

    @classmethod
    def from_document(cls, document: dict[str, Any]) -> Self:
        return cls(**document)

    @classmethod
    def from_raw_bson(cls, raw_bson) -> Any:
        pass

    def get(self, name, default=None) -> Any:
        return self.to_dict().get(name, default)

    def encode(self, **encode_kwargs: Any) -> bytes:
        raise NotImplementedError

    def decode(self, data: Any, **decode_kwargs: Any) -> Any:
        raise NotImplementedError

    def to_json_type(self) -> dict[str, Any]:
        raise NotImplementedError

    def to_dict(self) -> dict[str, Any]:
        raise NotImplementedError

    def to_tuple(self) -> tuple[Any, ...]:
        raise NotImplementedError

    # Operations
    def insert(self) -> None:
        """Insert this document"""
        # Send insert signal
        signal("insert").send(self.__class__, frames=[self])

        document_dict = self.to_dict()
        if not self._id:
            document_dict.pop("_id", None)
        # Prepare the document to be inserted
        document = to_refs(document_dict)

        # Insert the document and update the Id
        self._id = self.get_collection().insert_one(document).inserted_id

        # Send inserted signal
        signal("inserted").send(self.__class__, frames=[self])

    def unset(self, *fields: Any) -> None:
        """Unset the given list of fields for this document."""

        # Send update signal
        signal("update").send(self.__class__, frames=[self])

        # Clear the fields from the document and build the unset object
        unset = {}
        for field in fields:
            setattr(self, field, self._empty_type)
            unset[field] = True

        # Update the document
        self.get_collection().update_one({"_id": self._id}, {"$unset": unset})

        # Send updated signal
        signal("updated").send(self.__class__, frames=[self])

    def update(self, *fields: Any) -> None:
        """
        Update this document. Optionally a specific list of fields to update can
        be specified.
        """
        self_document = self.to_dict()
        assert "_id" in self_document, "Can't update documents without `_id`"

        # Send update signal
        signal("update").send(self.__class__, frames=[self])

        # Check for selective updates
        if len(fields) > 0:
            document = {}
            for field in fields:
                document[field] = self._path_to_value(field, self_document)
        else:
            document = self_document

        # Prepare the document to be updated
        document = to_refs(document)
        document.pop("_id", None)

        # Update the document
        self.get_collection().update_one({"_id": self._id}, {"$set": document})

        # Send updated signal
        signal("updated").send(self.__class__, frames=[self])

    def upsert(self, *fields: Any) -> None:
        """
        Update or Insert this document depending on whether it exists or not.
        The presense of an `_id` value in the document is used to determine if
        the document exists.

        NOTE: This method is not the same as specifying the `upsert` flag when
        calling MongoDB. When called for a document with an `_id` value, this
        method will call the database to see if a record with that Id exists,
        if not it will call `insert`, if so it will call `update`. This
        operation is therefore not atomic and much slower than the equivalent
        MongoDB operation (due to the extra call).
        """

        # If no `_id` is provided then we insert the document
        if not self._id:
            return self.insert()

        # If an `_id` is provided then we need to check if it exists before
        # performing the `upsert`.
        #
        if self.count({"_id": self._id}) == 0:
            self.insert()
        else:
            self.update(*fields)

    def delete(self) -> None:
        """Delete this document"""

        assert "_id" in self.to_dict(), "Can't delete documents without `_id`"

        # Send delete signal
        signal("delete").send(self.__class__, frames=[self])

        # Delete the document
        self.get_collection().delete_one({"_id": self._id})

        # Send deleted signal
        signal("deleted").send(self.__class__, frames=[self])

    @classmethod
    def find(cls, filter=None, **kwargs) -> list[Mapping[str, Any]]:
        """Return a list of documents matching the filter"""
        # Flatten the projection
        kwargs["projection"], references, subs = cls._flatten_projection(
            kwargs.get("projection", cls._default_projection)
        )

        # Find the document
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        documents = list(cls.get_collection().find(to_refs(filter), **kwargs))

        # Make sure we found documents
        if not documents:
            return []

        # Dereference the documents (if required)
        if references:
            cls._dereference(documents, references)

        # Add sub-frames to the documents (if required)
        if subs:
            cls._apply_sub_frames(documents, subs)

        return documents

    @classmethod
    def find_one(cls, filter=None, **kwargs) -> Mapping[str, Any]:
        """Return the first document matching the filter"""
        # Flatten the projection
        kwargs["projection"], references, subs = cls._flatten_projection(
            kwargs.get("projection", cls._default_projection)
        )

        # Find the document
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        document = cls.get_collection().find_one(to_refs(filter), **kwargs)

        # Make sure we found a document
        if not document:
            return {}

        # Dereference the document (if required)
        if references:
            cls._dereference([document], references)

        # Add sub-frames to the document (if required)
        if subs:
            cls._apply_sub_frames([document], subs)

        return document

    def reload(self, **kwargs):
        """Reload the document"""
        frame = self.find_one({"_id": self._id}, **kwargs)
        for field in frame:
            setattr(self, field, frame[field])

    @classmethod
    def insert_many(cls, documents: SpecsOrRawDocuments) -> Specs:
        """Insert a list of documents"""
        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        # Send insert signal
        signal("insert").send(cls, frames=frames)

        # Prepare the documents to be inserted
        _documents = [to_refs(f.to_dict()) for f in frames]

        for _document in _documents:
            if not _document["_id"]:
                _document.pop("_id")

        # Bulk insert
        ids = cls.get_collection().insert_many(_documents).inserted_ids

        # Apply the Ids to the frames
        for i, id in enumerate(ids):
            frames[i]._id = id

        # Send inserted signal
        signal("inserted").send(cls, frames=frames)

        return frames

    @classmethod
    def update_many(cls, documents: SpecsOrRawDocuments, *fields: Any) -> None:
        """
        Update multiple documents. Optionally a specific list of fields to
        update can be specified.
        """
        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        all_count = len(documents)
        assert len([f for f in frames if "_id" in f.to_dict()]) == all_count, "Can't update documents without `_id`s"

        # Send update signal
        signal("update").send(cls, frames=frames)

        # Prepare the documents to be updated

        # Check for selective updates
        if len(fields) > 0:
            _documents = []
            for frame in frames:
                document = {"_id": frame._id}
                for field in fields:
                    document[field] = cls._path_to_value(field, frame.to_dict())
                _documents.append(to_refs(document))
        else:
            _documents = [to_refs(f.to_dict()) for f in frames]

        # Update the documents
        requests = []
        for _document in _documents:
            _id = _document.pop("_id")
            requests.append(UpdateOne({"_id": _id}, {"$set": _document}))

        cls.get_collection().bulk_write(requests)

        # Send updated signal
        signal("updated").send(cls, frames=frames)

    @classmethod
    def unset_many(cls, documents: SpecsOrRawDocuments, *fields: Any) -> None:
        """Unset the given list of fields for given documents."""

        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        all_count = len(documents)
        assert len([f for f in frames if "_id" in f.to_dict()]) == all_count, "Can't update documents without `_id`s"

        # Send update signal
        signal("update").send(cls, frames=frames)

        # Clear the fields from the documents and build a list of ids to
        # update.
        ids = []
        for frame in frames:
            if frame._id:
                ids.append(frame._id)

        # Build the unset object
        unset = {}
        for field in fields:
            unset[field] = True
            for frame in frames:
                frame.to_dict().pop(field, None)

        # Update the document
        cls.get_collection().update_many({"_id": {"$in": ids}}, {"$unset": unset})

        # Send updated signal
        signal("updated").send(cls, frames=frames)

    @classmethod
    def delete_many(cls, documents: SpecsOrRawDocuments) -> None:
        """Delete multiple documents"""

        # Ensure all documents have been converted to frames
        frames = cls._ensure_frames(documents)

        all_count = len(documents)
        assert len([f for f in frames if "_id" in f.to_dict()]) == all_count, "Can't delete documents without `_id`s"

        # Send delete signal
        signal("delete").send(cls, frames=frames)

        # Prepare the documents to be deleted
        ids = [f._id for f in frames]

        # Delete the documents
        cls.get_collection().delete_many({"_id": {"$in": ids}})

        # Send deleted signal
        signal("deleted").send(cls, frames=frames)

    # Querying

    @classmethod
    def by_id(cls, id, **kwargs) -> Optional[Self]:
        """Get a document by ID"""
        return cls.one({"_id": id}, **kwargs)

    @classmethod
    def count(cls, filter=None, **kwargs) -> int:
        """Return a count of documents matching the filter"""
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        filter = to_refs(filter)

        if filter:
            return cls.get_collection().count_documents(to_refs(filter), **kwargs)
        else:
            return cls.get_collection().estimated_document_count(**kwargs)

    @classmethod
    def ids(cls, filter=None, **kwargs) -> list[ObjectId]:
        """Return a list of Ids for documents matching the filter"""
        # Find the documents
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        documents = cls.get_collection().find(to_refs(filter), projection={"_id": True}, **kwargs)

        return [d["_id"] for d in list(documents)]

    @classmethod
    def one(cls, filter=None, **kwargs) -> Optional[Self]:
        """Return the first spec object matching the filter"""
        # Flatten the projection
        kwargs["projection"], references, subs = cls._flatten_projection(
            kwargs.get("projection", cls._default_projection)
        )

        # Find the document
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        document = cls.get_collection().find_one(to_refs(filter), **kwargs)

        # Make sure we found a document
        if not document:
            return None

        # Dereference the document (if required)
        if references:
            cls._dereference([document], references)

        # Add sub-frames to the document (if required)
        if subs:
            cls._apply_sub_frames([document], subs)

        return cls.from_document(document)

    @classmethod
    def many(cls, filter=None, **kwargs) -> list[Self]:
        """Return a list of spec objects matching the filter"""
        # Flatten the projection
        kwargs["projection"], references, subs = cls._flatten_projection(
            kwargs.get("projection", cls._default_projection)
        )

        # Find the documents
        if isinstance(filter, (Condition, Group)):
            filter = filter.to_dict()

        documents = list(cls.get_collection().find(to_refs(filter), **kwargs))

        # Dereference the documents (if required)
        if references:
            cls._dereference(documents, references)

        # Add sub-frames to the documents (if required)
        if subs:
            cls._apply_sub_frames(documents, subs)

        return [cls(**d) for d in documents]

    @classmethod
    def get_collection(cls) -> Collection[Any]:
        """Return a reference to the database collection for the class"""
        if cls._collection_context is not None:
            return cls._collection_context

        return getattr(cls.get_db(), cls._collection or cls.__name__)

    @classmethod
    def get_db(cls) -> Database:
        """Return the database for the collection"""
        if not cls._client:
            raise NotImplementedError("_client is not setup yet")
        if cls._db is not None:
            return getattr(cls._client, cls._db.name)
        return cls._client.get_default_database()

    @classmethod
    @contextmanager
    def with_options(cls, **options: Any) -> Generator[Any, Any, None]:
        existing_context = getattr(cls, "_collection_context", None)

        try:
            collection = cls.get_collection()
            cls._collection_context = collection.with_options(**options)
            yield cls._collection_context

        finally:
            if cls._collection_context is None:
                del cls._collection_context
            else:
                cls._collection_context = existing_context

    @classmethod
    def _path_to_value(cls, path: str, parent_dict: dict[str, Any]) -> Any:
        """Return a value from a dictionary at the given path"""
        keys: list[str] = cls._path_to_keys(path)  # type: ignore

        # Traverse to the tip of the path
        child_dict = parent_dict
        for key in keys[:-1]:
            child_dict = child_dict.get(key)  # type: ignore[assignment]

            # unpaved path- return None
            if child_dict is None:
                return None

        return child_dict.get(keys[-1])

    @classmethod
    def _path_to_keys(cls, path: str) -> list[str]:
        """Return a list of keys for a given path"""
        return path.split(".")

    @classmethod
    def _ensure_frames(cls, documents: SpecsOrRawDocuments) -> Specs:
        """
        Ensure all items in a list are frames by converting those that aren't.
        """
        frames = []
        for document in documents:
            if not isinstance(document, cls):
                frames.append(cls(**document))
            else:
                frames.append(document)
        return frames

    @classmethod
    def _apply_sub_frames(cls, documents: RawDocuments, subs: dict[str, Any]) -> None:
        """Convert embedded documents to sub-frames for one or more documents"""

        # Dereference each reference
        for path, projection in subs.items():
            # Get the SubFrame class we'll use to wrap the embedded document
            sub = None
            expect_map = False
            if "$sub" in projection:
                sub = projection.pop("$sub")
            elif "$sub." in projection:
                sub = projection.pop("$sub.")
                expect_map = True
            else:
                continue

            # Add sub-frames to the documents
            raw_subs: list[Any] = []
            for document in documents:
                value = cls._path_to_value(path, document)
                if value is None:
                    continue

                if isinstance(value, dict):
                    if expect_map:
                        # Dictionary of embedded documents
                        raw_subs += value.values()
                        for k, v in value.items():
                            if isinstance(v, list):
                                value[k] = [sub(u) for u in v if isinstance(u, dict)]
                            else:
                                value[k] = sub(**v)

                    # Single embedded document
                    else:
                        raw_subs.append(value)
                        value = sub(**value)

                elif isinstance(value, list):
                    # List of embedded documents
                    raw_subs += value
                    value = [sub(**v) for v in value if isinstance(v, dict)]

                else:
                    raise TypeError("Not a supported sub-frame type")

                child_document = document
                keys = cls._path_to_keys(path)
                for key in keys[:-1]:
                    child_document = child_document[key]
                child_document[keys[-1]] = value

            # Apply the projection to the list of sub frames
            if projection:
                sub._apply_projection(raw_subs, projection)

    @classmethod
    def _flatten_projection(cls, projection: dict[str, Any]) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
        """
        Flatten a structured projection (structure projections support for
        projections of (to be) dereferenced fields.
        """

        # If `projection` is empty return a full projection based on `_fields`
        if not projection:
            return {f: True for f in cls.get_fields()}, {}, {}

        # Flatten the projection
        flat_projection = {}
        references = {}
        subs = {}
        inclusive = True
        for key, value in deepcopy(projection).items():
            if isinstance(value, dict):
                # Build the projection value for the field (allowing for
                # special mongo directives).
                values_to_project = {
                    k: v for k, v in value.items() if k.startswith("$") and k not in ["$ref", "$sub", "$sub."]
                }
                project_value = True if len(values_to_project) == 0 else {key: values_to_project}

                if project_value is not True:
                    inclusive = False

                # Store a reference/sub-frame projection
                if "$ref" in value:
                    references[key] = value

                elif "$sub" in value or "$sub." in value:
                    subs[key] = value
                    sub_frame = None
                    if "$sub" in value:
                        sub_frame = value["$sub"]

                    if "$sub." in value:
                        sub_frame = value["$sub."]

                    if sub_frame:
                        project_value = sub_frame._projection_to_paths(key, value)

                if isinstance(project_value, dict):
                    flat_projection.update(project_value)

                else:
                    flat_projection[key] = project_value

            elif key == "$ref":
                # Strip any $ref key
                continue

            elif key == "$sub" or key == "$sub.":
                # Strip any $sub key
                continue

            elif key.startswith("$"):
                # Strip mongo operators
                continue

            else:
                # Store the root projection value
                flat_projection[key] = value
                inclusive = False

        # If only references and sub-frames where specified in the projection
        # then return a full projection based on `_fields`.
        if inclusive:
            flat_projection = {f: True for f in cls.get_fields()}

        return flat_projection, references, subs

    @classmethod
    def _dereference(cls, documents: RawDocuments, references: dict[str, Any]):
        """Dereference one or more documents"""

        # Dereference each reference
        for path, projection in references.items():
            # Check there is a $ref in the projection, else skip it
            if "$ref" not in projection:
                continue

            # Collect Ids of documents to dereference
            ids = set()
            for document in documents:
                value = cls._path_to_value(path, document)
                if not value:
                    continue

                if isinstance(value, list):
                    ids.update(value)

                elif isinstance(value, dict):
                    ids.update(value.values())

                else:
                    ids.add(value)

            # Find the referenced documents
            ref = projection.pop("$ref")

            frames = ref.many({"_id": {"$in": list(ids)}}, projection=projection)
            frames = {f._id: f for f in frames}

            # Add dereferenced frames to the document
            for document in documents:
                value = cls._path_to_value(path, document)
                if not value:
                    continue

                if isinstance(value, list):
                    # List of references
                    value = [frames[id] for id in value if id in frames]

                elif isinstance(value, dict):
                    # Dictionary of references
                    value = {key: frames.get(id) for key, id in value.items()}

                else:
                    value = frames.get(value, None)

                child_document = document
                keys = cls._path_to_keys(path)
                for key in keys[:-1]:
                    child_document = child_document[key]
                child_document[keys[-1]] = value

    # Signals
    @classmethod
    def listen(cls, event: str, func: Callable) -> None:
        """Add a callback for a signal against the class"""
        signal(event).connect(func, sender=cls)

    @classmethod
    def stop_listening(cls, event: str, func: Callable) -> None:
        """Remove a callback for a signal against the class"""
        signal(event).disconnect(func, sender=cls)

    # Integrity helpers

    @classmethod
    def cascade(cls, ref_cls, field, frames) -> None:
        """Apply a cascading delete (does not emit signals)"""
        ids = [to_refs(getattr(f, field)) for f in frames if hasattr(f, field)]
        ref_cls.get_collection().delete_many({"_id": {"$in": ids}})

    @classmethod
    def nullify(cls, ref_cls, field, frames) -> None:
        """Nullify a reference field (does not emit signals)"""
        ids = [to_refs(f) for f in frames]
        ref_cls.get_collection().update_many({field: {"$in": ids}}, {"$set": {field: None}})

    @classmethod
    def pull(cls, ref_cls, field, frames) -> None:
        """Pull references from a list field (does not emit signals)"""
        ids = [to_refs(f) for f in frames]
        ref_cls.get_collection().update_many({field: {"$in": ids}}, {"$pull": {field: {"$in": ids}}})

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return self._id == other._id

    def __lt__(self, other: Any) -> Any:
        return self._id < other._id


class SubSpecBase:
    _parent: ClassVar[Any] = SpecBase

    def to_dict(self) -> Any:
        raise NotImplementedError()

    @classmethod
    def _apply_projection(cls, documents, projection):
        # Find reference and sub-frame mappings
        references = {}
        subs = {}
        for key, value in deepcopy(projection).items():
            if not isinstance(value, dict):
                continue

            # Store a reference/sub-frame projection
            if "$ref" in value:
                references[key] = value
            elif "$sub" in value or "$sub." in value:
                subs[key] = value

        # Dereference the documents (if required)
        if references:
            cls._parent._dereference(documents, references)

        # Add sub-frames to the documents (if required)
        if subs:
            cls._parent._apply_sub_frames(documents, subs)

    @classmethod
    def _projection_to_paths(cls, root_key, projection):
        """
        Expand a $sub/$sub. projection to a single projection of True (if
        inclusive) or a map of full paths (e.g `employee.company.tel`).
        """

        # Referenced projections are handled separately so just flag the
        # reference field to true.
        if "$ref" in projection:
            return True

        inclusive = True
        sub_projection = {}
        for key, value in projection.items():
            if key in ["$sub", "$sub."]:
                continue

            if key.startswith("$"):
                sub_projection[root_key] = {key: value}
                inclusive = False
                continue

            sub_key = root_key + "." + key

            if isinstance(value, dict):
                sub_value = cls._projection_to_paths(sub_key, value)
                if isinstance(sub_value, dict):
                    sub_projection.update(sub_value)
                else:
                    sub_projection[sub_key] = True

            else:
                sub_projection[sub_key] = True
                inclusive = False

        if inclusive:
            # No specific keys so this is inclusive
            return True

        return sub_projection


def to_refs(value: Any) -> Any:
    """Convert all Frame instances within the given value to Ids"""
    # Frame
    if isinstance(value, SpecBase):
        return getattr(value, "_id")

    # SubFrame
    elif isinstance(value, SubSpecBase):
        return to_refs(value.to_dict())

    # Lists
    elif isinstance(value, (list, tuple)):
        return [to_refs(v) for v in value]

    # Dictionaries
    elif isinstance(value, dict):
        return {k: to_refs(v) for k, v in value.items()}

    return value



from typing import Any

__all__ = ["deep_merge"]


def deep_merge(source: dict[str, Any], dest: dict[str, Any]) -> None:
    """
    Deep merges source dict into dest dict.

    This code was taken directly from the mongothon project:
    https://github.com/gamechanger/mongothon/tree/master/mongothon
    """
    for key, value in source.items():
        if key in dest:
            if isinstance(value, dict) and isinstance(dest[key], dict):
                deep_merge(value, dest[key])
                continue
            elif isinstance(value, list) and isinstance(dest[key], list):
                for item in value:
                    if item not in dest[key]:
                        dest[key].append(item)
                continue
        dest[key] = value



from pymongo import ASCENDING, DESCENDING

from .base import to_refs
from .query import Condition, Group
from .utils import deep_merge

__all__ = [
    # Operators
    "All",
    "ElemMatch",
    "Exists",
    "In",
    "Not",
    "NotIn",
    "Size",
    "Type",
    # Groups
    "And",
    "Or",
    "Nor",
    # Sorting
    "SortBy",
]

# Operators


def All(q, value):
    """
    The All operator selects documents where the value of the field is an list
    that contains all the specified elements.
    """
    return Condition(q._path, to_refs(value), "$all")


def ElemMatch(q, *conditions):
    """
    The ElemMatch operator matches documents that contain an array field with at
    least one element that matches all the specified query criteria.
    """
    new_condition = {}
    for condition in conditions:
        if isinstance(condition, (Condition, Group)):
            condition = condition.to_dict()

        deep_merge(condition, new_condition)

    return Condition(q._path, new_condition, "$elemMatch")


def Exists(q, value):
    """
    When exists is True, Exists matches the documents that contain the field,
    including documents where the field value is null. If exists is False, the
    query returns only the documents that do not contain the field.
    """
    return Condition(q._path, value, "$exists")


def In(q, value):
    """
    The In operator selects the documents where the value of a field equals any
    value in the specified list.
    """
    return Condition(q._path, to_refs(value), "$in")


def Not(condition):
    """
    Not performs a logical NOT operation on the specified condition and selects
    the documents that do not match. This includes documents that do not contain
    the field.
    """
    return Condition(condition.q, {condition.operator: condition.value}, "$not")


def NotIn(q, value):
    """
    The NotIn operator selects documents where the field value is not in the
    specified list or the field does not exists.
    """
    return Condition(q._path, to_refs(value), "$nin")


def Size(q, value):
    """
    The Size operator matches any list with the number of elements specified by
    size.
    """
    return Condition(q._path, value, "$size")


def Type(q, value):
    """
    Type selects documents where the value of the field is an instance of the
    specified BSON type.
    """
    return Condition(q._path, value, "$type")


# Groups


class And(Group):
    """
    And performs a logical AND operation on a list of two or more conditions and
    selects the documents that satisfy all the conditions.
    """

    operator = "$and"


class Or(Group):
    """
    The Or operator performs a logical OR operation on a list of two or more
    conditions and selects the documents that satisfy at least one of the
    conditions.
    """

    operator = "$or"


class Nor(Group):
    """
    Nor performs a logical NOR operation on a list of one or more conditions and
    selects the documents that fail all the conditions.
    """

    operator = "$nor"


# Sorting


def SortBy(*qs):
    """Convert a list of Q objects into list of sort instructions"""

    sort = []
    for q in qs:
        if q._path.endswith(".desc"):
            sort.append((q._path[:-5], DESCENDING))
        else:
            sort.append((q._path, ASCENDING))
    return sort



from pymongo import ASCENDING as ASC
from pymongo import DESCENDING as DESC
from pymongo.collation import Collation
from pymongo.operations import IndexModel

from .empty import Empty
from .ops import All, And, ElemMatch, Exists, In, Nor, Not, NotIn, Or, Size, SortBy, Type
from .pagination import Page, Paginator
from .query import Q
from .se import MongoDecoder, MongoEncoder

__all__ = [
    # Queries
    "Q",
    # Operators
    "All",
    "ElemMatch",
    "Exists",
    "In",
    "Not",
    "NotIn",
    "Size",
    "Type",
    # Groups
    "And",
    "Or",
    "Nor",
    # Sorting
    "SortBy",
    # Se
    "MongoEncoder",
    "MongoDecoder",
    # pymongo
    "Collation",
    "IndexModel",
    "ASC",
    "DESC",
    # Empty
    "Empty",
    # Pagination
    "Paginator",
    "Page",
]



from datetime import date
from typing import Any, Callable, Optional

import bson
import msgspec

from mongospecs.base import SpecBase

from .empty import Empty, EmptyObject


def bson_enc_hook(obj: Any) -> Any:
    if obj is msgspec.UNSET or obj is Empty:
        return None
    if type(obj) == date:
        return str(obj)
    if isinstance(obj, bson.ObjectId):
        return str(obj)

    raise NotImplementedError(f"Objects of type {type(obj)} are not supported")


def bson_dec_hook(typ, val) -> Any:
    if typ == bson.ObjectId:
        return bson.ObjectId(val)
    if typ == EmptyObject:
        return Empty


def encode(obj: Any, enc_hook: Optional[Callable[[Any], Any]] = bson_enc_hook) -> bytes:
    return bson.encode(msgspec.to_builtins(obj, enc_hook=enc_hook, builtin_types=(bson.ObjectId,)))


def encode_spec(obj: SpecBase) -> bytes:
    return bson.encode(obj.to_json_type())


def decode(msg: bytes, typ=Any, dec_hook: Optional[Callable[[type, Any], Any]] = bson_dec_hook):
    return msgspec.convert(bson.decode(msg), type=typ, dec_hook=dec_hook)



from typing import Annotated, Any, Callable, ClassVar, Optional, TypeVar, Union, cast

from blinker import signal
from bson import ObjectId
from pydantic import BaseModel, ConfigDict, Field
from pydantic_core import core_schema
from pymongo import MongoClient

from .base import EmptyObject, SpecBase, SubSpecBase

__all__ = ["Spec", "SubSpec"]


class _ObjectIdPydanticAnnotation:
    # Based on https://docs.pydantic.dev/latest/usage/types/custom/#handling-third-party-types.

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: Callable[[Any], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        def validate_from_str(input_value: str) -> ObjectId:
            return ObjectId(input_value)

        return core_schema.union_schema(
            [
                # check if it's an instance first before doing any further work
                core_schema.is_instance_schema(ObjectId),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ],
            serialization=core_schema.to_string_ser_schema(),
        )


PyObjectId = Annotated[ObjectId, _ObjectIdPydanticAnnotation]


class Spec(BaseModel, SpecBase):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    id: Optional[PyObjectId] = Field(default=None, alias="_id")

    @classmethod
    def from_document(cls, document: dict[str, Any]) -> "Spec":
        return cls.model_construct(**document)

    @property
    def _id(self) -> Union[EmptyObject, ObjectId]:
        return cast(Union[EmptyObject, ObjectId], self.id)

    @_id.setter
    def _id(self, value: ObjectId) -> None:
        self.id = value

    def unset(self, *fields: Any) -> None:
        """Unset the given list of fields for this document."""

        # Send update signal
        signal("update").send(self.__class__, frames=[self])

        # Clear the fields from the document and build the unset object
        unset = {}
        for field in fields:
            setattr(self, field, self._empty_type)
            unset[field] = True

            ## pydantic specific change:
            ## remove from model fields set so it excludes when `to_json_type` is called
            self.model_fields_set.remove(field)

        # Update the document
        self.get_collection().update_one({"_id": self._id}, {"$unset": unset})

        # Send updated signal
        signal("updated").send(self.__class__, frames=[self])

    def encode(self, **encode_kwargs: Any) -> bytes:
        return str.encode(self.model_dump_json(**encode_kwargs))

    def decode(self, data: Any, **decode_kwargs: Any) -> Any:
        return self.__class__.model_validate_json(data, **decode_kwargs)

    def to_json_type(self) -> Any:
        return self.model_dump(mode="json", by_alias=True, exclude_unset=True)

    def to_dict(self) -> dict[str, Any]:
        copy_dict = self.__dict__.copy()
        copy_dict["_id"] = copy_dict.pop("id")
        return copy_dict

    # note: pydantic BaseModel already has to_tuple

    @classmethod
    def get_fields(cls) -> set[str]:
        return set(cls.model_fields.keys())


class SubSpec(BaseModel, SubSpecBase):
    _parent: ClassVar[Any] = Spec

    def get(self, name: str, default: Any = None) -> Any:
        return self.to_dict().get(name, default)

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump()


class PydanticAdapter(Spec, BaseModel):
    def __init__(self, **data: Any) -> None:
        """Create a new model by parsing and validating input data from keyword arguments.

        Raises [ValidationError][pydantic_core.ValidationError] if the input data cannot
        be validated to form a valid model.

        __init__ uses __pydantic_self__ instead of the more common self for the first arg
        to allow self as a field name.
        """
        ...


T = TypeVar("T", bound=PydanticAdapter)


class AdapterBuilder:
    def __call__(
        self, obj: type[BaseModel], *, collection: str, client: Optional[MongoClient] = None, **kwds: Any
    ) -> Any:
        class BuiltSpecAdapter(obj, Spec):  # type: ignore
            pass

        BuiltSpecAdapter.__name__ = f"{obj.__name__}SpecAdapter"
        BuiltSpecAdapter._collection = collection
        BuiltSpecAdapter.__doc__ = obj.__doc__
        if client:
            BuiltSpecAdapter._client = client
        return BuiltSpecAdapter


SpecAdapter = AdapterBuilder()



from datetime import date, datetime
from typing import Any, ClassVar, Optional, TypeVar

import attrs
import msgspec
from attr import AttrsInstance
from bson import ObjectId
from pymongo import MongoClient

from .base import SpecBase, SpecProtocol, SubSpecBase
from .empty import Empty
from .se import MongoEncoder, mongo_dec_hook

__all__ = ["Spec", "SubSpec"]

T = TypeVar("T")


def attrs_serializer(inst: type, field: attrs.Attribute, value: Any) -> Any:
    if type(value) == date:
        return str(value)
    elif isinstance(value, ObjectId):
        return str(value)
    elif isinstance(value, datetime):
        return datetime.isoformat(value)
    return value


@attrs.define(kw_only=True)
class Spec(SpecBase):
    _id: Optional[ObjectId] = attrs.field(default=None, alias="_id", repr=True)  # type: ignore[assignment]

    @property
    def id(self) -> Optional[ObjectId]:
        return self._id

    @id.setter
    def id(self, value: ObjectId) -> None:
        self._id = value

    @classmethod
    def get_fields(cls) -> set[str]:
        return {f.name for f in attrs.fields(cls)}  # type: ignore[misc]

    def encode(self, **encode_kwargs: Any) -> bytes:
        return msgspec.json.encode(self, **encode_kwargs) if encode_kwargs else MongoEncoder.encode(self)

    def decode(self, data: Any, **decode_kwargs: Any) -> Any:
        return msgspec.json.decode(data, type=self.__class__, dec_hook=mongo_dec_hook, **decode_kwargs)

    def to_json_type(self) -> Any:
        return attrs.asdict(
            self,
            filter=lambda attr, value: value is not Empty or (attr.name == "_id" and value is not None),
            value_serializer=attrs_serializer,
        )

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self, recurse=False).copy()

    def to_tuple(self) -> tuple[Any, ...]:
        return attrs.astuple(self)


@attrs.define(kw_only=True)
class SubSpec(SubSpecBase):
    _parent: ClassVar[Any] = Spec

    def get(self, name, default=None):  # -> Any:
        return self.to_dict().get(name, default)

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)


class AttrsAdapter(AttrsInstance, SpecProtocol): ...


class AdapterBuilder:
    def __call__(
        self, obj: type[AttrsInstance], *, collection: str, client: Optional[MongoClient] = None, **kwds: Any
    ) -> Any:
        @attrs.define(kw_only=True)
        class BuiltSpecAdapter(Spec, obj):  # type: ignore
            ...

        BuiltSpecAdapter.__name__ = f"{obj.__name__}SpecAdapter"
        BuiltSpecAdapter._collection = collection
        BuiltSpecAdapter.__doc__ = obj.__doc__
        if client:
            BuiltSpecAdapter._client = client
        return BuiltSpecAdapter  # type: ignore


SpecAdapter = AdapterBuilder()



"""
Support for paginating specs.
"""

import math
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Generator, Generic, Optional, TypeVar, Union

from .base import SpecBase, to_refs
from .query import Condition, Group

__all__ = (
    # Exceptions
    "InvalidPage",
    # Classes
    "Page",
    "Paginator",
)


T = TypeVar("T")
TSpec = TypeVar("TSpec", bound=SpecBase)


class InvalidPage(Exception):
    """
    An error raised when an invalid page is requested.
    """


@dataclass
class Page(Generic[T]):
    """
    A class to represent one page of results.
    """

    offset: int
    """The offset of the first result in the page from the first result of
        the entire selection.
    """

    number: int
    """The page number"""

    items: list[T]
    """The results/specs for this page"""

    next: Optional[int]
    """Next page number"""

    prev: Optional[int]
    """Previous page number"""

    def __getitem__(self, i: int) -> T:
        return self.items[i]

    def __iter__(self):
        for item in self.items:
            yield item

    def __len__(self) -> int:
        return len(self.items)

    def get_offset(self, item: Any) -> int:
        """Return the offset for an item in the page"""
        return self.offset + self.items.index(item)


@dataclass
class Paginator(Generic[TSpec]):
    """
    A pagination class for slicing query results into pages. This class is
    designed to work with Spec classes.
    """

    spec_cls: type[TSpec]
    """The spec class results are being paginated for"""

    filter: Optional[Union[dict[str, Any], Condition, Group]] = None
    """The filter applied when selecting results from the database"""

    per_page: int = 20
    """The number of results that will be displayed per page"""

    orphans: int = 0
    """ If a value is specified for orphans then the last page will be able to
        hold the additional results (up to the value of orphans). This can
        help prevent users being presented with pages contain only a few
        results.
    """

    filter_kwargs: Any = None
    """Any additional filter arguments applied when selecting results such as sort and projection"""

    def __post_init__(self):
        # flattern the filter at this point which effectively deep copies.
        if isinstance(self.filter, (Condition, Group)):
            self.filter = self.filter.to_dict()
        else:
            self.filter = to_refs(self.filter)

        # Count the total results being paginated
        self._items_count = self.spec_cls.count(self.filter)

        # Calculated the number of pages
        total = self._items_count - self.orphans
        self._page_count = max(1, int(math.ceil(total / float(self.per_page))))

        # Create a list of page number that can be used to navigate the results
        self._page_numbers = range(1, self._page_count + 1)

    def __getitem__(self, page_number: int) -> Page[TSpec]:
        if page_number not in self._page_numbers:
            raise InvalidPage(page_number, self.page_count)

        # Calculate the next and previous page numbers
        next = page_number + 1 if page_number + 1 in self._page_numbers else None
        prev = page_number - 1 if page_number - 1 in self._page_numbers else None

        # Select the items for the page
        filter_args = deepcopy(self.filter_kwargs) or {}
        filter_args["skip"] = (page_number - 1) * self.per_page
        filter_args["limit"] = self.per_page

        # Check to see if we need to account for orphans
        if self.item_count - (page_number * self.per_page) <= self.orphans:
            filter_args["limit"] += self.orphans

        # Select the results for the page
        items = self.spec_cls.many(self.filter, **filter_args)

        # Build the page
        return Page(offset=filter_args["skip"], number=page_number, items=items, next=next, prev=prev)

    def __iter__(self) -> Generator[Page[TSpec], Any, Any]:
        for page_number in self._page_numbers:
            yield self[page_number]

    # Read-only properties
    @property
    def item_count(self) -> int:
        """Return the total number of items being paginated"""
        return self._items_count

    @property
    def page_count(self) -> int:
        """Return the total number of pages"""
        return self._page_count

    @property
    def page_numbers(self) -> range:
        """Return a list of page numbers"""
        return self._page_numbers



from datetime import date
from typing import Any

try:
    import attrs

    ATTRS_NOTHING_TYPE = attrs.NOTHING
except ImportError:
    ATTRS_NOTHING_TYPE = type("BaseNothing")  # type: ignore[assignment]

import msgspec
from bson import ObjectId

__all__ = ["MongoEncoder", "MongoDecoder"]


def mongo_enc_hook(obj: Any) -> Any:
    if obj is msgspec.UNSET or obj is ATTRS_NOTHING_TYPE:
        return None
    elif type(obj) == date:
        return str(obj)
    # Object Id
    elif isinstance(obj, ObjectId):
        return str(obj)

    raise NotImplementedError(f"Objects of type {type(obj)} are not supported")


def mongo_dec_hook(typ: Any, obj: Any) -> Any:
    if typ is ObjectId:
        return ObjectId(obj)
    raise NotImplementedError(f"Objects of type {type(obj)} are not supported")


MongoEncoder = msgspec.json.Encoder(enc_hook=mongo_enc_hook)
MongoDecoder = msgspec.json.Decoder(dec_hook=mongo_dec_hook)
